@@ -975,6 +975,8 @@ sub setup ($) {
 	    DEFAULT => {
 		"pre-build-commands" => [],
 		"chroot-setup-commands" => [],
+		"starting-build-commands" => [],
+		"finished-build-commands" => [],
 		"chroot-cleanup-commands" => [],
 		"post-build-commands" => [],
 	    },
@@ -989,6 +991,14 @@ sub setup ($) {
         [\'foo\', \'arg1\', \'arg2\'],
         [\'bar\', \'arg1\', \'arg2\', \'arg3\'],
     ],
+    "starting-build-commands" => [
+        [\'foo\', \'arg1\', \'arg2\'],
+        [\'bar\', \'arg1\', \'arg2\', \'arg3\'],
+    ],
+    "finished-build-commands" => [
+        [\'foo\', \'arg1\', \'arg2\'],
+        [\'bar\', \'arg1\', \'arg2\', \'arg3\'],
+    ],
     "chroot-cleanup-commands" => [
         [\'foo\', \'arg1\', \'arg2\'],
         [\'bar\', \'arg1\', \'arg2\', \'arg3\'],
@@ -1049,6 +1059,18 @@ sub setup ($) {
 	    DEFAULT => '/var/lib/sbuild/apt-keys/sbuild-key.pub',
 	    HELP => 'GPG public key for temporary local apt archive.'
 	},
+	'EXTRA_PACKAGES'				=> {
+	    TYPE => 'ARRAY:STRING',
+	    GROUP => '__INTERNAL',
+	    DEFAULT => [],
+	    HELP => 'Additional per-build packages available as build dependencies.  Do not set by hand.'
+	},
+	'EXTRA_REPOSITORIES'				=> {
+	    TYPE => 'ARRAY:STRING',
+	    GROUP => '__INTERNAL',
+	    DEFAULT => [],
+	    HELP => 'Additional per-build apt repositories.  Do not set by hand.'
+	},
     );
 
     $conf->set_allowed_keys(\%sbuild_keys);
@@ -1097,7 +1119,7 @@ if (\%individual_stalled_pkg_timeout) {
 END
 
     my $custom_setup = <<END;
-push(\@{\${\$conf->get('EXTERNAL_COMMANDS')}{"chroot-setup-commands"}},
+push(\@{\${\$conf->get('EXTERNAL_COMMANDS')}{"chroot-user-setup-commands"}},
 \$chroot_setup_script) if (\$chroot_setup_script);
 
     # Trigger log directory creation if needed
